@@ -4,14 +4,19 @@
 #include <gdk/gdkx.h>
 #include <gdk/gdkkeysyms.h>
 #include "eggcellrendererkeys.h"
+#include "eggaccelerators.h"
 
+#ifndef EGG_COMPILATION
 #ifndef _
 #define _(x) dgettext (GETTEXT_PACKAGE, x)
 #define N_(x) x
 #endif
+#else
+#define _(x) x
+#define N_(x) x
+#endif
 
 #define EGG_CELL_RENDERER_TEXT_PATH "egg-cell-renderer-text"
-#define USED_MODS (GDK_SHIFT_MASK | GDK_CONTROL_MASK | GDK_MOD1_MASK)
 
 static void             egg_cell_renderer_keys_finalize      (GObject             *object);
 static void             egg_cell_renderer_keys_init          (EggCellRendererKeys *cell_keys);
@@ -25,21 +30,29 @@ static GtkCellEditable *egg_cell_renderer_keys_start_editing (GtkCellRenderer
 							      GtkCellRendererState      flags);
 
 
-static void egg_cell_renderer_keys_get_property  (GObject                  *object,
-						  guint                     param_id,
-						  GValue                   *value,
-						  GParamSpec               *pspec);
-static void egg_cell_renderer_keys_set_property  (GObject                  *object,
-						  guint                     param_id,
-						  const GValue             *value,
-						  GParamSpec               *pspec);
+static void egg_cell_renderer_keys_get_property (GObject         *object,
+						 guint            param_id,
+						 GValue          *value,
+						 GParamSpec      *pspec);
+static void egg_cell_renderer_keys_set_property (GObject         *object,
+						 guint            param_id,
+						 const GValue    *value,
+						 GParamSpec      *pspec);
+static void egg_cell_renderer_keys_get_size     (GtkCellRenderer *cell,
+						 GtkWidget       *widget,
+						 GdkRectangle    *cell_area,
+						 gint            *x_offset,
+						 gint            *y_offset,
+						 gint            *width,
+						 gint            *height);
+
 
 enum {
   PROP_0,
 
-  /* FIXME make names consistent with something else */
   PROP_ACCEL_KEY,
-  PROP_ACCEL_MASK
+  PROP_ACCEL_MASK,
+  PROP_ACCEL_MODE
 };
 
 static GtkCellRendererTextClass *parent_class = NULL;
@@ -59,7 +72,7 @@ egg_cell_renderer_keys_get_type (void)
         (GClassInitFunc)egg_cell_renderer_keys_class_init,
 	NULL,		/* class_finalize */
 	NULL,		/* class_data */
-        sizeof (GtkCellRendererText),
+        sizeof (EggCellRendererKeys),
 	0,              /* n_preallocs */
         (GInstanceInitFunc) egg_cell_renderer_keys_init
       };
@@ -73,28 +86,30 @@ egg_cell_renderer_keys_get_type (void)
 static void
 egg_cell_renderer_keys_init (EggCellRendererKeys *cell_keys)
 {
+  cell_keys->accel_mode = EGG_CELL_RENDERER_KEYS_MODE_GTK;
 }
 
 /* FIXME setup stuff to generate this */
-/* VOID:STRING,UINT,FLAGS */
+/* VOID:STRING,UINT,FLAGS,UINT */
 static void
-marshal_VOID__STRING_UINT_FLAGS (GClosure     *closure,
-                                 GValue       *return_value,
-                                 guint         n_param_values,
-                                 const GValue *param_values,
-                                 gpointer      invocation_hint,
-                                 gpointer      marshal_data)
+marshal_VOID__STRING_UINT_FLAGS_UINT (GClosure     *closure,
+                                      GValue       *return_value,
+				      guint         n_param_values,
+				      const GValue *param_values,
+				      gpointer      invocation_hint,
+				      gpointer      marshal_data)
 {
-  typedef void (*GMarshalFunc_VOID__STRING_UINT_FLAGS) (gpointer     data1,
-                                                        const char  *arg_1,
-                                                        guint        arg_2,
-                                                        int          arg_3,
-                                                        gpointer     data2);
-  register GMarshalFunc_VOID__STRING_UINT_FLAGS callback;
+  typedef void (*GMarshalFunc_VOID__STRING_UINT_FLAGS_UINT) (gpointer     data1,
+                                                             const char  *arg_1,
+							     guint        arg_2,
+							     int          arg_3,
+							     guint        arg_4,
+							     gpointer     data2);
+  register GMarshalFunc_VOID__STRING_UINT_FLAGS_UINT callback;
   register GCClosure *cc = (GCClosure*) closure;
   register gpointer data1, data2;
 
-  g_return_if_fail (n_param_values == 4);
+  g_return_if_fail (n_param_values == 5);
 
   if (G_CCLOSURE_SWAP_DATA (closure))
     {
@@ -107,12 +122,13 @@ marshal_VOID__STRING_UINT_FLAGS (GClosure     *closure,
       data2 = closure->data;
     }
   
-  callback = (GMarshalFunc_VOID__STRING_UINT_FLAGS) (marshal_data ? marshal_data : cc->callback);
+  callback = (GMarshalFunc_VOID__STRING_UINT_FLAGS_UINT) (marshal_data ? marshal_data : cc->callback);
 
   callback (data1,
             g_value_get_string (param_values + 1),
             g_value_get_uint (param_values + 2),
             g_value_get_flags (param_values + 3),
+	    g_value_get_uint (param_values + 4),
             data2);
 }
 
@@ -120,15 +136,17 @@ static void
 egg_cell_renderer_keys_class_init (EggCellRendererKeysClass *cell_keys_class)
 {
   GObjectClass *object_class;
-  
-  object_class = G_OBJECT_CLASS (cell_keys_class);
+  GtkCellRendererClass *cell_renderer_class;
 
+  object_class = G_OBJECT_CLASS (cell_keys_class);
+  cell_renderer_class = GTK_CELL_RENDERER_CLASS (cell_keys_class);
   parent_class = g_type_class_peek_parent (object_class);
   
   GTK_CELL_RENDERER_CLASS (cell_keys_class)->start_editing = egg_cell_renderer_keys_start_editing;
 
   object_class->set_property = egg_cell_renderer_keys_set_property;
   object_class->get_property = egg_cell_renderer_keys_get_property;
+  cell_renderer_class->get_size = egg_cell_renderer_keys_get_size;
 
   object_class->finalize = egg_cell_renderer_keys_finalize;
   
@@ -136,15 +154,6 @@ egg_cell_renderer_keys_class_init (EggCellRendererKeysClass *cell_keys_class)
    * to match whatever the GTK convention is
    */
   
-  g_object_class_install_property (object_class,
-                                   PROP_ACCEL_MASK,
-                                   g_param_spec_flags ("accel_mask",
-                                                       _("Accelerator modifiers"),
-                                                       _("Accelerator modifiers"),
-                                                       GDK_TYPE_MODIFIER_TYPE,
-                                                       0,
-                                                       G_PARAM_READABLE | G_PARAM_WRITABLE));
-  
   g_object_class_install_property (object_class,
                                    PROP_ACCEL_KEY,
                                    g_param_spec_uint ("accel_key",
@@ -155,16 +164,37 @@ egg_cell_renderer_keys_class_init (EggCellRendererKeysClass *cell_keys_class)
                                                       0,
                                                       G_PARAM_READABLE | G_PARAM_WRITABLE));
 
+  g_object_class_install_property (object_class,
+                                   PROP_ACCEL_MASK,
+                                   g_param_spec_flags ("accel_mask",
+                                                       _("Accelerator modifiers"),
+                                                       _("Accelerator modifiers"),
+                                                       GDK_TYPE_MODIFIER_TYPE,
+                                                       0,
+                                                       G_PARAM_READABLE | G_PARAM_WRITABLE));
+  
+  /* FIXME: Register the enum when moving to GTK+ */
+  g_object_class_install_property (object_class,
+                                   PROP_ACCEL_MODE,
+                                   g_param_spec_int ("accel_mode",
+						     _("Accelerator Mode"),
+						     _("The type of accelerator."),
+						     0,
+						     2,
+						     0,
+						     G_PARAM_READABLE | G_PARAM_WRITABLE));
+  
   g_signal_new ("keys_edited",
                 EGG_TYPE_CELL_RENDERER_KEYS,
                 G_SIGNAL_RUN_LAST,
                 G_STRUCT_OFFSET (EggCellRendererKeysClass, keys_edited),
                 NULL, NULL,
-                marshal_VOID__STRING_UINT_FLAGS,
-                G_TYPE_NONE, 3,
+                marshal_VOID__STRING_UINT_FLAGS_UINT,
+                G_TYPE_NONE, 4,
                 G_TYPE_STRING,
                 G_TYPE_UINT,
-                GDK_TYPE_MODIFIER_TYPE);
+                GDK_TYPE_MODIFIER_TYPE,
+		G_TYPE_UINT);
 }
 
 
@@ -182,13 +212,13 @@ egg_cell_renderer_keys_finalize (GObject *object)
 }
 
 static gchar *
-convert_keysym_state_to_string (guint           keysym,
-                                GdkModifierType state)
+convert_keysym_state_to_string (guint                  keysym,
+                                EggVirtualModifierType mask)
 {
   if (keysym == 0)
     return g_strdup (_("Disabled"));
   else
-    return gtk_accelerator_name (keysym, state);
+    return egg_virtual_accelerator_name (keysym, mask);
 }
 
 static void
@@ -213,6 +243,10 @@ egg_cell_renderer_keys_get_property  (GObject                  *object,
       g_value_set_flags (value, keys->accel_mask);
       break;
 
+    case PROP_ACCEL_MODE:
+      g_value_set_int (value, keys->accel_mode);
+      break;
+
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
     }
@@ -244,6 +278,10 @@ egg_cell_renderer_keys_set_property  (GObject                  *object,
                                               g_value_get_flags (value));
       break;
 
+    case PROP_ACCEL_MODE:
+      egg_cell_renderer_keys_set_accel_mode (keys, g_value_get_int (value));
+      break;
+      
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
     }
@@ -276,63 +314,133 @@ is_modifier (guint keycode)
   return retval;
 }
 
-static GdkFilterReturn
-grab_key_filter (GdkXEvent *gdk_xevent, GdkEvent *event, gpointer data)
+void
+egg_cell_renderer_keys_get_size (GtkCellRenderer *cell,
+				 GtkWidget       *widget,
+				 GdkRectangle    *cell_area,
+				 gint            *x_offset,
+				 gint            *y_offset,
+				 gint            *width,
+				 gint            *height)
 {
-  XEvent *xevent = (XEvent *)gdk_xevent;
-  guint keycode, state;
-  char buf[10];
-  KeySym keysym;
+  EggCellRendererKeys *keys = (EggCellRendererKeys *) cell;
+  GtkRequisition requisition;
+
+  if (keys->sizing_label == NULL)
+    keys->sizing_label = gtk_label_new (_("Type a new accelerator, or press Backspace to clear"));
+
+  gtk_widget_size_request (keys->sizing_label, &requisition);
+  (* GTK_CELL_RENDERER_CLASS (parent_class)->get_size) (cell, widget, cell_area, x_offset, y_offset, width, height);
+  /* FIXME: need to take the cell_area et al. into account */
+  if (width)
+    *width = MAX (*width, requisition.width);
+  if (height)
+    *height = MAX (*height, requisition.height);
+}
+
+/* FIXME: Currently we don't differentiate between a 'bogus' key (like tab in
+ * GTK mode) and a removed key.
+ */
+   
+static gboolean
+grab_key_callback (GtkWidget    *widget,
+                   GdkEventKey  *event,
+                   void         *data)
+{
+  GdkModifierType accel_mods;
+  guint accel_keyval;
   EggCellRendererKeys *keys;
   char *path;
   gboolean edited;
+  GdkModifierType consumed_modifiers;  
+  guint upper;
+  GdkModifierType ignored_modifiers;
   
-  if (xevent->type != KeyPress)
-    return GDK_FILTER_CONTINUE;
-	
   keys = EGG_CELL_RENDERER_KEYS (data);
-  
-  keycode = xevent->xkey.keycode;
+  accel_mods = 0;
 
-  if (is_modifier (keycode))
-    return GDK_FILTER_CONTINUE;
+  if (is_modifier (event->hardware_keycode))
+    return TRUE;
 
   edited = FALSE;
+
+  consumed_modifiers = 0;
+  gdk_keymap_translate_keyboard_state (gdk_keymap_get_default (),
+				       event->hardware_keycode,
+                                       event->state,
+                                       event->group,
+				       NULL, NULL, NULL, &consumed_modifiers);
+
+  upper = event->keyval;
+  accel_keyval = gdk_keyval_to_lower (upper);
+  if (accel_keyval == GDK_ISO_Left_Tab) 
+    accel_keyval = GDK_Tab;
+
+
   
-  state = xevent->xkey.state & USED_MODS;
+  /* Put shift back if it changed the case of the key, not otherwise.
+   */
+  if (upper != accel_keyval &&
+      (consumed_modifiers & GDK_SHIFT_MASK))
+    {
+      consumed_modifiers &= ~(GDK_SHIFT_MASK);
+    }
+
+  egg_keymap_resolve_virtual_modifiers (gdk_keymap_get_default (),
+                                        EGG_VIRTUAL_NUM_LOCK_MASK |
+                                        EGG_VIRTUAL_SCROLL_LOCK_MASK,
+                                        &ignored_modifiers);
   
-  XLookupString ((XKeyEvent*)xevent, buf, sizeof (buf), &keysym, NULL);
-  
-  if (state == 0 && keysym == GDK_Escape)
+  /* filter consumed/ignored modifiers */
+
+  if (keys->accel_mode == EGG_CELL_RENDERER_KEYS_MODE_GTK)
+    accel_mods = event->state & ~(consumed_modifiers | ignored_modifiers);
+  else if (keys->accel_mode == EGG_CELL_RENDERER_KEYS_MODE_X)
+    accel_mods = event->state & ~(ignored_modifiers);
+  else
+    g_assert_not_reached ();
+    
+  if (accel_mods == 0 && accel_keyval == GDK_Escape)
     goto out; /* cancel */
 
   /* clear the accelerator on Backspace */
   if (keys->edit_key != 0 &&
-      state == 0 &&
-      keysym == GDK_BackSpace)
-    keysym = 0;
+      accel_mods == 0 &&
+      accel_keyval == GDK_BackSpace)
+    accel_keyval = 0;
 
-  edited = TRUE;
+  if (keys->accel_mode == EGG_CELL_RENDERER_KEYS_MODE_GTK)
+    {
+      if (!gtk_accelerator_valid (accel_keyval, accel_mods))
+	{
+	  accel_keyval = 0;
+	  accel_mods = 0;
+	}
+
+      /* Remove modifiers like super and hyper, as GTK+ ignores them. */
+      accel_mods = accel_mods & GDK_MODIFIER_MASK;
+    }
   
+  edited = TRUE;
  out:
   path = g_strdup (g_object_get_data (G_OBJECT (keys->edit_widget),
                                       EGG_CELL_RENDERER_TEXT_PATH));
   
-  gdk_keyboard_ungrab (xevent->xkey.time);
-  gdk_pointer_ungrab (xevent->xkey.time);
+  gdk_keyboard_ungrab (event->time);
+  gdk_pointer_ungrab (event->time);
   
   gtk_cell_editable_editing_done (GTK_CELL_EDITABLE (keys->edit_widget));
   gtk_cell_editable_remove_widget (GTK_CELL_EDITABLE (keys->edit_widget));
   keys->edit_widget = NULL;
-  keys->filter_window = NULL;
+  keys->grab_widget = NULL;
   
   if (edited)
     g_signal_emit_by_name (G_OBJECT (keys), "keys_edited", path,
-                           keysym, state);
+                           accel_keyval, accel_mods, event->hardware_keycode);
 
   g_free (path);
   
-  return GDK_FILTER_REMOVE;
+  return TRUE;
 }
 
 static void
@@ -343,13 +451,13 @@ ungrab_stuff (GtkWidget *widget, gpointer data)
   gdk_keyboard_ungrab (GDK_CURRENT_TIME);
   gdk_pointer_ungrab (GDK_CURRENT_TIME);
 
-  gdk_window_remove_filter (keys->filter_window,
-			    grab_key_filter, data);
+  g_signal_handlers_disconnect_by_func (G_OBJECT (keys->grab_widget),
+                                        G_CALLBACK (grab_key_callback), data);
 }
 
 static void
 pointless_eventbox_start_editing (GtkCellEditable *cell_editable,
-                               GdkEvent        *event)
+                                  GdkEvent        *event)
 {
   /* do nothing, because we are pointless */
 }
@@ -430,9 +538,11 @@ egg_cell_renderer_keys_start_editing (GtkCellRenderer      *cell,
       return NULL;
     }
   
-  keys->filter_window = widget->window;
-  
-  gdk_window_add_filter (keys->filter_window, grab_key_filter, keys);
+  keys->grab_widget = widget;
+
+  g_signal_connect (G_OBJECT (widget), "key_press_event",
+                    G_CALLBACK (grab_key_callback),
+                    keys);
 
   eventbox = g_object_new (pointless_eventbox_subclass_get_type (),
                            NULL);
@@ -450,8 +560,8 @@ egg_cell_renderer_keys_start_editing (GtkCellRenderer      *cell,
                         &widget->style->fg[GTK_STATE_SELECTED]);
   
   if (keys->accel_key != 0)
-    gtk_label_set_markup (GTK_LABEL (label),
-                          _("Type a new accelerator, or press Backspace to clear"));
+    gtk_label_set_text (GTK_LABEL (label),
+			_("Type a new accelerator, or press Backspace to clear"));
   else
     gtk_label_set_text (GTK_LABEL (label),
                         _("Type a new accelerator"));
@@ -467,22 +577,21 @@ egg_cell_renderer_keys_start_editing (GtkCellRenderer      *cell,
                     G_CALLBACK (ungrab_stuff), keys);
   
   keys->edit_key = keys->accel_key;
-  keys->edit_mask = keys->accel_mask;
   
   return GTK_CELL_EDITABLE (keys->edit_widget);
 }
 
 void
-egg_cell_renderer_keys_set_accelerator (EggCellRendererKeys *keys,
-                                        guint                keyval,
-                                        GdkModifierType      mask)
+egg_cell_renderer_keys_set_accelerator (EggCellRendererKeys    *keys,
+                                        guint                   keyval,
+                                        EggVirtualModifierType  mask)
 {
   char *text;
   gboolean changed;
   GtkCellRendererText *celltext;
-  
+
   g_return_if_fail (EGG_IS_CELL_RENDERER_KEYS (keys));
-  
+
   g_object_freeze_notify (G_OBJECT (keys));
 
   changed = FALSE;
@@ -497,29 +606,27 @@ egg_cell_renderer_keys_set_accelerator (EggCellRendererKeys *keys,
   if (mask != keys->accel_mask)
     {
       keys->accel_mask = mask;
+
       g_object_notify (G_OBJECT (keys), "accel_mask");
       changed = TRUE;
     }  
+  g_object_thaw_notify (G_OBJECT (keys));
 
   if (changed)
     {
       /* sync string to the key values */
       celltext = GTK_CELL_RENDERER_TEXT (keys);
-      
       text = convert_keysym_state_to_string (keys->accel_key, keys->accel_mask);
-      g_free (celltext->text);
-      celltext->text = text;
-      g_object_notify (G_OBJECT (keys), "text");
+      g_object_set (keys, "text", text, NULL);
+      g_free (text);
     }
   
-  g_object_thaw_notify (G_OBJECT (keys));
 }
 
-
 void
-egg_cell_renderer_keys_get_accelerator (EggCellRendererKeys *keys,
-                                        guint               *keyval,
-                                        GdkModifierType     *mask)
+egg_cell_renderer_keys_get_accelerator (EggCellRendererKeys     *keys,
+                                        guint                   *keyval,
+                                        EggVirtualModifierType  *mask)
 {
   g_return_if_fail (EGG_IS_CELL_RENDERER_KEYS (keys));
 
@@ -529,3 +636,12 @@ egg_cell_renderer_keys_get_accelerator (EggCellRendererKeys *keys,
   if (mask)
     *mask = keys->accel_mask;
 }
+
+void
+egg_cell_renderer_keys_set_accel_mode (EggCellRendererKeys     *keys,
+				       EggCellRendererKeysMode  accel_mode)
+{
+  g_return_if_fail (EGG_IS_CELL_RENDERER_KEYS (keys));
+  keys->accel_mode = accel_mode;
+  g_object_notify (G_OBJECT (keys), "accel_mode");
+}
