@@ -907,6 +907,35 @@ remove_from_list_callback (GtkObject *object, gpointer data)
   *listp = g_slist_remove (*listp, object);
 }
 
+static gboolean
+cb_check_for_uniqueness (GtkTreeModel *model,
+                         GtkTreePath  *path,
+                         GtkTreeIter  *iter,
+                         gpointer      user_data)
+{
+  KeyEntry *key_entry;
+  KeyEntry *tmp_key_entry;
+ 
+  key_entry = (KeyEntry *) user_data;
+  gtk_tree_model_get (model, iter,
+                      KEYVAL_COLUMN, &tmp_key_entry,
+                      -1);
+ 
+  if (tmp_key_entry != NULL &&
+      key_entry->gconf_keyval == tmp_key_entry->gconf_keyval &&
+      key_entry->gconf_mask   == tmp_key_entry->gconf_mask &&
+      /* be sure we don't claim a key is a dup of itself */
+      strcmp (key_entry->gconf_key, tmp_key_entry->gconf_key) != 0)
+    {
+      key_entry->needs_gconf_sync = FALSE;
+      key_entry->gconf_key = tmp_key_entry->gconf_key;
+      key_entry->user_visible_name = tmp_key_entry->user_visible_name;
+      return TRUE;
+    }
+
+  return FALSE;
+}
+
 static void
 accel_edited_callback (GtkCellRendererText *cell,
                        const char          *path_string,
@@ -915,13 +944,16 @@ accel_edited_callback (GtkCellRendererText *cell,
                        guint                hardware_keycode,
                        gpointer             data)
 {
-  GtkTreeModel *model = (GtkTreeModel *)data;
+  GtkTreeView  *view = (GtkTreeView *) data;
+  GtkTreeModel *model;
   GtkTreePath *path = gtk_tree_path_new_from_string (path_string);
   GtkTreeIter iter;
-  KeyEntry *ke;
+  KeyEntry *ke, tmp_key;
   GError *err;
   char *str;
   
+  model = gtk_tree_view_get_model (view);
+
   gtk_tree_model_get_iter (model, &iter, path);
   gtk_tree_model_get (model, &iter, KEYVAL_COLUMN, &ke, -1);
 
@@ -929,6 +961,45 @@ accel_edited_callback (GtkCellRendererText *cell,
   if (ke == NULL)
     return;
 
+  tmp_key.gconf_keyval = keyval;
+  tmp_key.gconf_mask = mask;
+  tmp_key.gconf_key = ke->gconf_key;
+  tmp_key.user_visible_name = NULL;
+  tmp_key.needs_gconf_sync = TRUE; /* kludge: we'll use this as return flag in the _foreach call */
+
+  if (keyval != 0) 
+    {
+      gtk_tree_model_foreach (model, cb_check_for_uniqueness, &tmp_key);
+
+      if (!tmp_key.needs_gconf_sync)
+        {
+          GtkWidget *dialog;
+          char *name;
+
+          name = egg_virtual_accelerator_name (keyval, mask);
+
+          dialog =
+            gtk_message_dialog_new (GTK_WINDOW (gtk_widget_get_toplevel (GTK_WIDGET (view))),
+                                    GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_MODAL,
+                                    GTK_MESSAGE_WARNING,
+                                    GTK_BUTTONS_OK,
+                                    _("The shortcut key \"%s\" is already bound to the \"%s\" action"),
+                                    name,
+                                    tmp_key.user_visible_name ? tmp_key.user_visible_name : tmp_key.gconf_key);
+          g_free (name);
+
+          gtk_dialog_run (GTK_DIALOG (dialog));
+          gtk_widget_destroy (dialog);
+
+          /* set it back to its previous value. */
+          egg_cell_renderer_keys_set_accelerator (EGG_CELL_RENDERER_KEYS (cell),
+                                                  ke->gconf_keyval, ke->gconf_mask);
+          gtk_tree_path_free (path);
+
+          return;
+        }
+    }
+
   str = binding_name (keyval, mask, FALSE);
 
   D (g_print ("Edited keyval %s, setting gconf to %s\n",
@@ -1136,7 +1207,7 @@ terminal_edit_keys_dialog_new (GtkWindow *transient_parent)
 				NULL);
   g_signal_connect (G_OBJECT (cell_renderer), "keys_edited",
                     G_CALLBACK (accel_edited_callback),
-                    tree);
+                    w);
   
   g_object_set (G_OBJECT (cell_renderer),
                 "editable", TRUE,
